/* $Id: netwrite.c,v 1.1 2002/06/15 15:53:33 jirka Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define BUFFSIZE 1<<14

void usage(void)
{
    fprintf(stderr,
        "Usage: netwrite [udp] <options> <host> <port>\n"
        "\n"
        "Default TCP protocol can be changed to UDP by ``udp'' argument.\n"
        "UDP options\n"
        "  currently none\n"
        "TCP options\n"
        "  -f          firewall mode, connetion is initiated by netread\n"
        "              host specification is ignored and can be omited\n"
        "general options\n"
        "  -v          verbose\n"
        "  -V          version\n");
    exit(1);
}

int main(int argc, char **argv)
{
    int udp = 0;
    int opt;
    int verbose = 0;
    int firewall = 0;
    char *host_str = NULL;
    struct hostent *host;
    char *service;
    struct servent *se;
    unsigned short port;
    struct sockaddr_in sin;
    int sock;
    char buf[BUFFSIZE];
    ssize_t count;

    if (argc < 3)
        usage();

    if (strcmp(argv[1], "udp") == 0) {
        udp = 1;
        argv[1] = argv[0];
    }

    while ((opt = getopt(argc - udp, argv + udp, "fvV")) != -1) {
        switch (opt) {
        case 'f':           /* firewall mode */
            if (udp)
                usage();
            firewall = 1;
            break;
        case 'v':           /* verbose */
            verbose = 1;
            break;
        case 'V':           /* version */
            printf("netrw tools version %s\n$Id: netwrite.c,v 1.1 2002/06/15 15:53:33 jirka Exp $\n", VERSION);
            return 0;
            break;
        case '?':           /* bad option */
            usage();
        }
    }

    if ((optind + udp) >= argc)
        usage();

    if (firewall) {
        if ((optind + udp + 1) < argc)
            service = argv[optind + udp + 1];
        else
            service = argv[optind + udp];
        if (service == NULL)
            usage();
    }
    else {
        if ((optind + udp + 1) >= argc
            || (host_str = argv[optind + udp]) == NULL
            || (service = argv[optind + udp + 1]) == NULL)
            usage();

        if (!(host = gethostbyname(host_str))) {
            perror("gethostbyname()");
            return 1;
        }
    }

    if ((se = getservbyname(service, (udp ? "udp" : "tcp"))))
        port = se->s_port;
    else {
        /* service may also be a port number */
        char *endptr = NULL;
        port = htons(strtol(service, &endptr, 0));
        if (*(service) == '\0' || *endptr != '\0') {
            fprintf(stderr, "%s/%s: no such service\n",
                    (udp ? "udp" : "tcp"), service);
            return 1;
        }
    }

    sin.sin_family = AF_INET;
    sin.sin_port = port;

    /* create socket */
    if (udp) {
        sin.sin_addr = **(struct in_addr **)host->h_addr_list;

        if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
            perror("sokcet()");
            return 1;
        }
    }
    else if (firewall) {
        int listen_sock;
        struct sockaddr_in r_addr;
        unsigned int addrsize = sizeof(r_addr);

        if ((listen_sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
            perror("sokcet()");
            return 1;
        }

        sin.sin_addr.s_addr = INADDR_ANY;
        if (bind(listen_sock, &sin, sizeof(sin))) {
            perror("bind()");
            return 1;
        }
        if (listen(listen_sock, 0)) {
            perror("listen()");
            return 1;
        }
        if (verbose)
            fprintf(stderr, "Listening on %d\n", ntohs(port));

        if ((sock = accept(listen_sock, &r_addr, &addrsize)) == -1) {
            perror("accept()");
            return 1;
        }
        if (verbose) {
            fprintf(stderr, "Accepted connection from %s:%d\n",
                    inet_ntoa(r_addr.sin_addr), ntohs(r_addr.sin_port));
        }
        close(listen_sock);
    }
    else {
        sin.sin_addr = **(struct in_addr **)host->h_addr_list;

        if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
            perror("sokcet()");
            return 1;
        }

        if (connect(sock, &sin, sizeof(sin))) {
            perror("connect()");
            return 1;
        }

        if (verbose) {
            fprintf(stderr, "Connected to %s:%d\n",
                    inet_ntoa(sin.sin_addr), ntohs(sin.sin_port));
        }
    }

#define write_sock(sock, buf, count, addr, size) \
    (udp ? sendto(sock, buf, count, 0, addr, size) \
         : write(sock, buf, count))

    while ((count = read(STDIN_FILENO, buf, BUFFSIZE)) > 0) {
        int wr;
        char *p;

        for (p = buf;
             (wr = write_sock(sock, p, count, &sin, sizeof(sin))) > 0;
             p += wr) {
            if (!(count -= wr))
                break;
        }
        if (wr < 0) {
            perror("write_sock()");
            return 1;
        }
    }
    if (count < 0) {
        perror("read()");
        return 1;
    }
  
    close(sock);
    return 0;
}

